'''
This test module validates the behavior of the BE generation translating the
bounded semantics of LTL as defined in :mod:`pynusmv_tools.bmcLTL.ast`
'''

from unittest                    import TestCase
from tests                       import utils as tests
from pynusmv.be.expression       import Be 
from pynusmv_tools.bmcLTL                import ast
from pynusmv_tools.bmcLTL.parsing        import parseLTL

from pynusmv.bmc                 import ltlspec
from pynusmv.bmc                 import utils as bmcutils
from pynusmv.parser              import parse_ltl_spec
from pynusmv.node                import Node

class TestSemantics(TestCase):
    
    def test_successor(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # when time is smaller than bound
            self.assertEqual(4, ast.successor(3, 5, 2))
            # when time reaches the bound and needs to jump back to the beginning
            # of the loop.
            self.assertEqual(2, ast.successor(5, 5, 2))
    
    def test_loop_condition(self):
        with tests.Configure(self, __file__, "/example.smv"):
            enc   = self.enc
            a     = enc.by_name['a']
            b     = enc.by_name['b']
            
            # self looping on zero
            _,k,l = 0,0,0
            loop_cond = ast.loop_condition(enc, k, l)
            # note: frozen and ivar are not taken into account
            manual    = a.at_time[0].boolean_expression.iff(a.at_time[0].boolean_expression) &\
                        b.at_time[0].boolean_expression.iff(b.at_time[0].boolean_expression)
            nusmv     = bmcutils.loop_condition(enc, k, l)
                    
            s_tool   = tests.canonical_cnf(loop_cond)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
            
            # simple k-l loop 5 - 0
            _,k,l = 0,5,0
            loop_cond = ast.loop_condition(enc, k, l)
            
            manual    = a.at_time[5].boolean_expression.iff(a.at_time[0].boolean_expression) &\
                        b.at_time[5].boolean_expression.iff(b.at_time[0].boolean_expression)
            nusmv     = bmcutils.loop_condition(enc, k, l)
                    
            s_tool   = tests.canonical_cnf(loop_cond)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
            
            # simple k-l loop 5 - 0
            _,k,l = 0,5,2
            loop_cond = ast.loop_condition(enc, k, l)
            
            manual    = a.at_time[5].boolean_expression.iff(a.at_time[2].boolean_expression) &\
                        b.at_time[5].boolean_expression.iff(b.at_time[2].boolean_expression)
            nusmv     = bmcutils.loop_condition(enc, k, l)
                    
            s_tool   = tests.canonical_cnf(loop_cond)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
        
    
    def test_constant_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Constant("TRUE")
            self.assertEqual(Be.true(self.mgr), expr.semantic_no_loop(self.enc, 0, 5))
            
            expr = ast.Constant("FALSE")
            self.assertEqual(Be.false(self.mgr), expr.semantic_no_loop(self.enc, 0, 5))
        
    def test_constant_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Constant("TRUE")
            self.assertEqual(Be.true(self.mgr), expr.semantic_with_loop(self.enc, 0, 5, 2))
            
            expr = ast.Constant("FALSE")
            self.assertEqual(Be.false(self.mgr), expr.semantic_with_loop(self.enc, 0, 5, 2))
        
    def test_variable_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Proposition("a")
            self.assertEqual(
                    self.enc.by_name['a'].at_time[3].boolean_expression, 
                    expr.semantic_no_loop(self.enc, 3, 5))
            # bound has little impact on the BE for a variable
            self.assertEqual(
                    self.enc.by_name['a'].at_time[3].boolean_expression, 
                    expr.semantic_no_loop(self.enc, 3, 3))
    
    def test_variable_withloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # bound and loop have no impact on the way BE for a variable is 
            # generated (successor index not enforced)
            expr = ast.Proposition("a")
            self.assertEqual(
                    self.enc.by_name['a'].at_time[3].boolean_expression, 
                    expr.semantic_with_loop(self.enc, 3, 5, 2))
            self.assertEqual(
                    self.enc.by_name['a'].at_time[3].boolean_expression, 
                    expr.semantic_with_loop(self.enc, 3, 3, 3))
    
    def test_not_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # constant
            expr = ast.Not(ast.Constant("TRUE"))
            self.assertEqual(
                ast.Constant("FALSE").semantic_no_loop(self.enc, 2, 3),
                expr.semantic_no_loop(self.enc, 2, 3))
            
            # variable
            expr = ast.Not(ast.Proposition("a"))
            self.assertEqual(
                -ast.Proposition("a").semantic_no_loop(self.enc, 2, 3),
                expr.semantic_no_loop(self.enc, 2, 3))
            
            # not
            expr = ast.Not(ast.Not(ast.Proposition("a")))
            self.assertEqual(
                ast.Proposition("a").semantic_no_loop(self.enc, 2, 3),
                expr.semantic_no_loop(self.enc, 2, 3))
            
            # expression: and
            expr = ast.Not(ast.And(ast.Proposition("a"), ast.Proposition("b")))
            self.assertEqual(
                -( ast.Proposition("a").semantic_no_loop(self.enc, 2, 3) 
                 & ast.Proposition("b").semantic_no_loop(self.enc, 2, 3)),
                expr.semantic_no_loop(self.enc, 2, 3))
            
            # expression: weak until
            expr = ast.WeakUntil(ast.Proposition("a"), ast.Proposition("b"))
            nega = ast.Not(expr)
            self.assertEqual(
                -expr.semantic_no_loop(self.enc, 2, 3),
                 nega.semantic_no_loop(self.enc, 2, 3))
    
    def test_not_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # constant
            expr = ast.Not(ast.Constant("TRUE"))
            self.assertEqual(
                ast.Constant("FALSE").semantic_with_loop(self.enc, 2, 3, 1),
                expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            # variable
            expr = ast.Not(ast.Proposition("a"))
            self.assertEqual(
                -ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1),
                expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            # not
            expr = ast.Not(ast.Not(ast.Proposition("a")))
            self.assertEqual(
                ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1),
                expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            # expression: and
            expr = ast.Not(ast.And(ast.Proposition("a"), ast.Proposition("b")))
            self.assertEqual(
                -( ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1) 
                 & ast.Proposition("b").semantic_with_loop(self.enc, 2, 3, 1)),
                expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            # expression: weak until
            expr = ast.WeakUntil(ast.Proposition("a"), ast.Proposition("b"))
            nega = ast.Not(expr)
            self.assertEqual(
                -expr.semantic_with_loop(self.enc, 2, 3, 1),
                 nega.semantic_with_loop(self.enc, 2, 3, 1))
        
    def test_and_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.And(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_no_loop(self.enc, 2, 3)
                    & ast.Proposition("b").semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
            
            expr = ast.And(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_no_loop(self.enc, 2, 3)
                    & ast.Eventually(ast.Proposition("b")).semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
    
    def test_and_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.And(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1)
                    & ast.Proposition("b").semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            expr = ast.And(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_with_loop(self.enc, 2, 3, 1)
                    & ast.Eventually(ast.Proposition("b")).semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
            
        
    def test_or_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Or(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_no_loop(self.enc, 2, 3)
                    | ast.Proposition("b").semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
            
            expr = ast.Or(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_no_loop(self.enc, 2, 3)
                    | ast.Eventually(ast.Proposition("b")).semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
    
    def test_or_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Or(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1)
                    | ast.Proposition("b").semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            expr = ast.Or(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_with_loop(self.enc, 2, 3, 1)
                    | ast.Eventually(ast.Proposition("b")).semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
        
    def test_xor_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Xor(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_no_loop(self.enc, 2, 3)
                    ^ ast.Proposition("b").semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
            
            expr = ast.Xor(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_no_loop(self.enc, 2, 3)
                    ^ ast.Eventually(ast.Proposition("b")).semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
    
    def test_xor_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Xor(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1)
                    ^ ast.Proposition("b").semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            expr = ast.Xor(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_with_loop(self.enc, 2, 3, 1)
                    ^ ast.Eventually(ast.Proposition("b")).semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
        
    def test_imply_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Imply(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_no_loop(self.enc, 2, 3).imply(
                      ast.Proposition("b").semantic_no_loop(self.enc, 2, 3))),
                    expr.semantic_no_loop(self.enc, 2, 3))
            
            expr = ast.Imply(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    (-ast.Globally(ast.Proposition("a")).semantic_no_loop(self.enc, 2, 3)
                    | ast.Eventually(ast.Proposition("b")).semantic_no_loop(self.enc, 2, 3)),
                    expr.semantic_no_loop(self.enc, 2, 3))
    
    def test_imply_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Imply(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    (-ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1)
                    | ast.Proposition("b").semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            expr = ast.Imply(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    (-ast.Globally(ast.Proposition("a")).semantic_with_loop(self.enc, 2, 3, 1)
                    | ast.Eventually(ast.Proposition("b")).semantic_with_loop(self.enc, 2, 3, 1)),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
        
    def test_iff_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Equiv(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_no_loop(self.enc, 2, 3).iff(
                      ast.Proposition("b").semantic_no_loop(self.enc, 2, 3))),
                    expr.semantic_no_loop(self.enc, 2, 3))
            
            expr = ast.Equiv(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_no_loop(self.enc, 2, 3).iff(
                      ast.Eventually(ast.Proposition("b")).semantic_no_loop(self.enc, 2, 3))),
                    expr.semantic_no_loop(self.enc, 2, 3))
        
    def test_iff_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            expr = ast.Equiv(ast.Proposition("a"), ast.Proposition("b"))
            self.assertEqual(
                    ( ast.Proposition("a").semantic_with_loop(self.enc, 2, 3, 1).iff(
                      ast.Proposition("b").semantic_with_loop(self.enc, 2, 3, 1))),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
            
            expr = ast.Equiv(ast.Globally(ast.Proposition("a")), ast.Eventually(ast.Proposition("b")))
            self.assertEqual(
                    ( ast.Globally(ast.Proposition("a")).semantic_with_loop(self.enc, 2, 3, 1).iff(
                      ast.Eventually(ast.Proposition("b")).semantic_with_loop(self.enc, 2, 3, 1))),
                    expr.semantic_with_loop(self.enc, 2, 3, 1))
        
    def test_until_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            enc  = self.enc
            a    = ast.Proposition("a")
            b    = ast.Proposition("b")
    
            expr = ast.Until(a, b)
            tool = expr.semantic_no_loop(enc, 0, 2)
            
            manual = ( b.semantic_no_loop(enc, 0, 2) 
                   | (a.semantic_no_loop(enc, 0, 2) & (b.semantic_no_loop(enc, 1, 2)
                   | (a.semantic_no_loop(enc, 1, 2) & (b.semantic_no_loop(enc, 2, 2))))))
            
            
            spec  = Node.from_ptr(parse_ltl_spec("a U b"))
            nusmv = ltlspec.bounded_semantics(self.befsm, spec, bound=2, loop=bmcutils.no_loopback())
            
            # normalized string representation of the BE's (make them comparable)
            s_tool  = tests.canonical_cnf(tool)
            s_nusmv = tests.canonical_cnf(nusmv)
            s_manual= tests.canonical_cnf(manual)
            
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
        
    def test_until_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            enc  = self.enc
            i,k,l= 0,2,0
            a    = ast.Proposition("a")
            b    = ast.Proposition("b")
    
            expr = ast.Until(a, b)
            tool = expr.semantic_with_loop(enc, i,k,l)
            
            manual = b.semantic_with_loop(enc, i, k, l) | \
                        (a.semantic_with_loop(enc, i, k, l) & b.semantic_with_loop(enc, i+1, k, l))
            
            
            spec  = Node.from_ptr(parse_ltl_spec("a U b"))
            nusmv = ltlspec.bounded_semantics(self.befsm, spec, bound=k, loop=l)
            
            tool   &= bmcutils.loop_condition(enc, k, l)
            manual &= bmcutils.loop_condition(enc, k, l)
            
            # normalized string representation of the BE's (make them comparable)
            s_tool  = tests.canonical_cnf(tool)
            s_nusmv = tests.canonical_cnf(nusmv)
            s_manual= tests.canonical_cnf(manual)
            
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
        
    def test_weak_until_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            enc  = self.enc
            a    = ast.Proposition("a")
            b    = ast.Proposition("b")
    
            expr = ast.WeakUntil(a, b)
            tool = expr.semantic_no_loop(enc, 0, 2)
            
            manual = ( b.semantic_no_loop(enc, 0, 2) 
                   | (a.semantic_no_loop(enc, 0, 2) & (b.semantic_no_loop(enc, 1, 2)
                   | (a.semantic_no_loop(enc, 1, 2) & (b.semantic_no_loop(enc, 2, 2))))))
            
            
            spec  = Node.from_ptr(parse_ltl_spec("a U b"))
            nusmv = ltlspec.bounded_semantics(self.befsm, spec, bound=2, loop=bmcutils.no_loopback())
            
            # normalized string representation of the BE's (make them comparable)
            s_tool  = tests.canonical_cnf(tool)
            s_nusmv = tests.canonical_cnf(nusmv)
            s_manual= tests.canonical_cnf(manual)
            
            self.assertEqual(s_tool, s_manual)
            self.assertEqual(s_tool, s_nusmv)
        
    def test_weak_until_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            enc  = self.enc
            mgr  = enc.manager
            i,k,l= 0,2,0
            a    = ast.Proposition("a")
            b    = ast.Proposition("b")
    
            expr = ast.WeakUntil(a, b)
            tool = expr.semantic_with_loop(enc, i,k,l)
            
            manual = b.semantic_with_loop(enc, i, k, l) | \
                        (a.semantic_with_loop(enc, i, k, l) & \
                            (b.semantic_with_loop(enc, i+1, k, l) | \
                                a.semantic_with_loop(enc, i+1, k, l) & Be.true(mgr)
                            )
                         )
            
            # normalized string representation of the BE's (make them comparable)
            s_tool  = tests.canonical_cnf(tool)
            s_manual= tests.canonical_cnf(manual)
            
            self.assertEqual(s_tool, s_manual)
        
    def test_globally_no_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            i,k     = 0,2
            enc     = self.enc
            mgr     = enc.manager
            a       = ast.Proposition("a")
            formula = ast.Globally(a)
            
            tool = formula.semantic_no_loop(enc, i, k)
            self.assertEqual(Be.false(mgr), tool)
        
    def test_globally_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            i,k,l   = 0,2,0
            enc     = self.enc
            a       = ast.Proposition("a")
            formula = ast.Globally(a)
            tool    = formula.semantic_with_loop(enc, i, k, l) &\
                        bmcutils.loop_condition(enc, k, l)
            
            nusmv   = ltlspec.bounded_semantics(
                            self.befsm, Node.from_ptr(parse_ltl_spec("G a")), k, l)
            
            manual  = a.semantic_with_loop(enc, i, k, l)   &\
                      a.semantic_with_loop(enc, i+1, k, l) &\
                      bmcutils.loop_condition(enc, k, l)
            
            # normalized string representation of the BE's (make them comparable)
            s_tool  = tests.canonical_cnf(tool)
            s_nusmv = tests.canonical_cnf(nusmv)
            s_manual= tests.canonical_cnf(manual)
            
            self.assertEqual(s_nusmv,  s_tool)
            self.assertEqual(s_manual, s_tool)
        
    def test_eventually_no_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            i,k     = 0,2
            enc     = self.enc
            a       = ast.Proposition("a")
            formula = ast.Eventually(a)
            
            tool    = formula.semantic_no_loop(enc, i, k)
            
            manual  = a.semantic_no_loop(enc, i+2, k)   |\
                      a.semantic_no_loop(enc, i+1, k) |\
                      a.semantic_no_loop(enc, i, k) 
            
            nusmv   = ltlspec.bounded_semantics(
                        self.befsm, Node.from_ptr(parse_ltl_spec("F a")), 
                        bound = k, 
                        loop  = bmcutils.no_loopback())
            
            # normalized string representation of the BE's (make them comparable)
            s_tool  = tests.canonical_cnf(tool)
            s_nusmv = tests.canonical_cnf(nusmv)
            s_manual= tests.canonical_cnf(manual)
            
            self.assertEqual(s_nusmv,  s_tool)
            self.assertEqual(s_manual, s_tool)
        
    def test_eventually_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            i,k,l   = 0,2,0
            enc     = self.enc
            a       = ast.Proposition("a")
            formula = ast.Eventually(a)
            
            tool    = formula.semantic_with_loop(enc, i, k, l)
            
            manual  = a.semantic_with_loop(enc, i+1, k, l) |\
                      a.semantic_with_loop(enc, i  , k, l) 
            
            nusmv   = ltlspec.bounded_semantics(
                        self.befsm, Node.from_ptr(parse_ltl_spec("F a")), 
                        bound = k, 
                        loop  = l)
            
            # normalized string representation of the BE's (make them comparable)
            loop_cond = bmcutils.loop_condition(enc, k, l)
            s_tool  = tests.canonical_cnf(tool   & loop_cond)
            s_manual= tests.canonical_cnf(manual & loop_cond)
            s_nusmv = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_nusmv,  s_tool)
            self.assertEqual(s_manual, s_tool)
        
    def test_next_noloop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            i,k     = 0,2
            enc     = self.enc
            # One step
            a       = ast.Proposition("a")
            formula = ast.Next(a)
            tool    = formula.semantic_no_loop(enc, i, k)
            manual  = a.semantic_no_loop(enc, 1, k)
            nusmv   = ltlspec.bounded_semantics(self.befsm, 
                                                Node.from_ptr(parse_ltl_spec("X a")),
                                                bound = k, 
                                                loop  = bmcutils.no_loopback())
            
            s_tool   = tests.canonical_cnf(tool)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
            
            # two steps
            formula = ast.Next(ast.Next(a))
            tool    = formula.semantic_no_loop(enc, i, k)
            manual  = a.semantic_no_loop(enc, 2, k)
            nusmv   = ltlspec.bounded_semantics(self.befsm, 
                                                Node.from_ptr(parse_ltl_spec("X X a")),
                                                bound = k, 
                                                loop  = bmcutils.no_loopback())
            
            s_tool   = tests.canonical_cnf(tool)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
            
            # Three steps (getting over k)
            formula = ast.Next(ast.Next(ast.Next(a)))
            tool    = formula.semantic_no_loop(enc, i, k)
            manual  = Be.false(enc.manager)
            nusmv   = ltlspec.bounded_semantics(self.befsm, 
                                                Node.from_ptr(parse_ltl_spec("X X X a")),
                                                bound = k, 
                                                loop  = bmcutils.no_loopback())
            
            s_tool   = tests.canonical_cnf(tool)
            s_manual = tests.canonical_cnf(manual)
            s_nusmv  = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
        
    def test_next_with_loop(self):
        with tests.Configure(self, __file__, "/example.smv"):
            i,k,l   = 0,2,0
            enc     = self.enc
            
            # One step
            a       = ast.Proposition("a")
            formula = ast.Next(a)
            tool    = formula.semantic_with_loop(enc, i, k, l)
            manual  = a.semantic_with_loop(enc, 1, k, l)
            nusmv   = ltlspec.bounded_semantics(self.befsm, 
                                                Node.from_ptr(parse_ltl_spec("X a")),
                                                bound = k, 
                                                loop  = l)
            
            loop_cond = bmcutils.loop_condition(enc, k, l)
            s_tool  = tests.canonical_cnf(tool   & loop_cond)
            s_manual= tests.canonical_cnf(manual & loop_cond)
            s_nusmv = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
            
            # two steps
            formula = ast.Next(ast.Next(a))
            tool    = formula.semantic_with_loop(enc, i, k, l)
            manual  = a.semantic_with_loop(enc, 0, k, l)
            nusmv   = ltlspec.bounded_semantics(self.befsm, 
                                                Node.from_ptr(parse_ltl_spec("X X a")),
                                                bound = k, 
                                                loop  = l)
            
            loop_cond = bmcutils.loop_condition(enc, k, l)
            s_tool  = tests.canonical_cnf(tool   & loop_cond)
            s_manual= tests.canonical_cnf(manual & loop_cond)
            s_nusmv = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
             
            # Three steps (getting over k)
            formula = ast.Next(ast.Next(ast.Next(a)))
            tool    = formula.semantic_with_loop(enc, i, k, l)
            manual  = a.semantic_with_loop(enc, 1, k, l)
            nusmv   = ltlspec.bounded_semantics(self.befsm, 
                                                Node.from_ptr(parse_ltl_spec("X X X a")),
                                                bound = k, 
                                                loop  = l)
            
            loop_cond = bmcutils.loop_condition(enc, k, l)
            s_tool  = tests.canonical_cnf(tool   & loop_cond)
            s_manual= tests.canonical_cnf(manual & loop_cond)
            s_nusmv = tests.canonical_cnf(nusmv)
            
            self.assertEqual(s_tool, s_nusmv)
            self.assertEqual(s_tool, s_manual)
    
    
    def validate_bounded_semantics(self, bound, custom_text, nusmv_text):
        fsm     = self.befsm
        # formulae
        formula = parseLTL(custom_text)
        fml_node= Node.from_ptr(parse_ltl_spec(nusmv_text))
        
        tool = formula.bounded_semantics(fsm, bound)
        smv  = ltlspec.bounded_semantics(fsm, fml_node, bound)
        # canonical forms        
        s_tool  = tests.canonical_cnf(tool)
        s_smv   = tests.canonical_cnf(smv)
         
        self.assertEqual(s_tool, s_smv)
      
    def test_bounded_semantics(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # length 0
            self.validate_bounded_semantics(0, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_bounded_semantics(0, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_bounded_semantics(0, "(a U b)", "(a U b)")
            self.validate_bounded_semantics(0, "a => () b", "a -> (X b)")
            
            # length 1
            self.validate_bounded_semantics(1, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_bounded_semantics(1, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_bounded_semantics(1, "(a U b)", "(a U b)")
            self.validate_bounded_semantics(1, "a => () b", "a -> (X b)")
            
            # length 2
            self.validate_bounded_semantics(1, "<>(a <=> !b)", "F (a <-> !b)")
            self.validate_bounded_semantics(1, "[](a <=> !b)", "G (a <-> !b)")
            self.validate_bounded_semantics(1, "(a U b)", "(a U b)")
            self.validate_bounded_semantics(1, "a => () b", "a -> (X b)")
        
    def test_nnf_constant(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula = parseLTL("TRUE")
            self.assertEqual("Constant(TRUE)",  str(formula.nnf(False)))
            self.assertEqual("Constant(FALSE)", str(formula.nnf(True)))
        
    def test_nnf_variable(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula = parseLTL("a")
            self.assertEqual("Proposition(a)",       str(formula.nnf(False)))
            self.assertEqual("(Not Proposition(a))", str(formula.nnf(True)))
        
    def test_nnf_not(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula = parseLTL("!a")
            self.assertEqual("Proposition(a)",       str(formula.nnf(True)))
            self.assertEqual("(Not Proposition(a))", str(formula.nnf(False)))
    
    def test_nnf_and(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # when the operands dont need to be nnf-ed
            formula = parseLTL("a & b")
    
            negated = "((Not Proposition(a)) Or (Not Proposition(b)))"
            self.assertEqual(negated,       str(formula.nnf(True)))
            
            not_neg = "(Proposition(a) And Proposition(b))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            # when the operands need to be nnf-ed
            formula = parseLTL("a & !b")
    
            negated = "((Not Proposition(a)) Or Proposition(b))"
            self.assertEqual(negated,       str(formula.nnf(True)))
            
            not_neg = "(Proposition(a) And (Not Proposition(b)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
    
    def test_nnf_or(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # when the operands dont need to be nnf-ed
            formula = parseLTL("a | b")
    
            negated = "((Not Proposition(a)) And (Not Proposition(b)))"
            self.assertEqual(negated,       str(formula.nnf(True)))
            
            not_neg = "(Proposition(a) Or Proposition(b))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            # when the operands need to be nnf-ed
            formula = parseLTL("a | !b")
    
            negated = "((Not Proposition(a)) And Proposition(b))"
            self.assertEqual(negated,       str(formula.nnf(True)))
            
            not_neg = "(Proposition(a) Or (Not Proposition(b)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
        
    def test_nnf_imply(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # when the operands dont need to be nnf-ed
            formula = parseLTL("a => b")
            
            not_neg = "((Not Proposition(a)) Or Proposition(b))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "(Proposition(a) And (Not Proposition(b)))"
            self.assertEqual(negated, str(formula.nnf(True)))
            
            # when the operands need to be nnf-ed
            formula = parseLTL("a => !b")
            
            not_neg = "((Not Proposition(a)) Or (Not Proposition(b)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "(Proposition(a) And Proposition(b))"
            self.assertEqual(negated, str(formula.nnf(True)))
    
    def test_nnf_equiv(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # when the operands dont need to be nnf-ed
            formula = parseLTL("a <=> b")
            
            #          (!a | b) & (!b | a)
            not_neg = "(((Not Proposition(a)) Or Proposition(b)) And ((Not Proposition(b)) Or Proposition(a)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            #          (a & !b) | (b & !a)
            negated = "((Proposition(a) And (Not Proposition(b))) Or (Proposition(b) And (Not Proposition(a))))"
            self.assertEqual(negated, str(formula.nnf(True)))
             
            # when the operands need to be nnf-ed
            formula = parseLTL("a <=> !b")
            
            #          (!a | !b) & (b | a)
            not_neg = "(((Not Proposition(a)) Or (Not Proposition(b))) And (Proposition(b) Or Proposition(a)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            #          (a & b) | (!b & !a) 
            negated = "((Proposition(a) And Proposition(b)) Or ((Not Proposition(b)) And (Not Proposition(a))))"
            self.assertEqual(negated, str(formula.nnf(True)))
    
    def test_nnf_until(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # rules of pseudo duality:
            #     (p U q) <-> (!q W (!p & !q))
            #     (p W q) <-> (!q U (!p & !q))
            
            # without operands nnfing
            formula = parseLTL("(p U q)")
            not_neg = "(Proposition(p) Until Proposition(q))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "((Not Proposition(q)) WeakUntil ((Not Proposition(p)) And (Not Proposition(q))))"
            self.assertEqual(negated, str(formula.nnf(True)))
            
            # wit operands nnfing
            formula = parseLTL("((!p) U (!q))")
            not_neg = "((Not Proposition(p)) Until (Not Proposition(q)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "(Proposition(q) WeakUntil (Proposition(p) And Proposition(q)))"
            self.assertEqual(negated, str(formula.nnf(True)))
                
    def test_nnf_weakuntil(self):
        with tests.Configure(self, __file__, "/example.smv"):
            # rules of pseudo duality:
            #     (p U q) <-> (!q W (!p & !q))
            #     (p W q) <-> (!q U (!p & !q))
            
            # without operands nnfing
            formula = parseLTL("(p W q)")
            not_neg = "(Proposition(p) WeakUntil Proposition(q))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "((Not Proposition(q)) Until ((Not Proposition(p)) And (Not Proposition(q))))"
            self.assertEqual(negated, str(formula.nnf(True)))
            
            # wit operands nnfing
            formula = parseLTL("((!p) W (!q))")
            not_neg = "((Not Proposition(p)) WeakUntil (Not Proposition(q)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "(Proposition(q) Until (Proposition(p) And Proposition(q)))"
            self.assertEqual(negated, str(formula.nnf(True)))
        
    def test_nnf_globally(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula = parseLTL("[] a")
            not_neg = "(Globally Proposition(a))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
    
            negated = "(Eventually (Not Proposition(a)))"
            self.assertEqual(negated, str(formula.nnf(True)))
            
            # when members need nnfing too
            formula = parseLTL("[] (!a)")
            not_neg = "(Globally (Not Proposition(a)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
    
            negated = "(Eventually Proposition(a))"
            self.assertEqual(negated, str(formula.nnf(True)))

    def test_nnf_eventually(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula = parseLTL("<> a")
            not_neg = "(Eventually Proposition(a))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
    
            negated = "(Globally (Not Proposition(a)))"
            self.assertEqual(negated, str(formula.nnf(True)))
            
            # when members need nnfing too
            formula = parseLTL("<> (!a)")
            not_neg = "(Eventually (Not Proposition(a)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
    
            negated = "(Globally Proposition(a))"
            self.assertEqual(negated, str(formula.nnf(True)))
    
    def test_nnf_next(self):
        with tests.Configure(self, __file__, "/example.smv"):
            formula = parseLTL("() a")
            # w/o member nnf-ing
            not_neg = "(Next Proposition(a))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "(Next (Not Proposition(a)))"
            self.assertEqual(negated, str(formula.nnf(True)))
            
            # w/ member nnf-ing
            formula = parseLTL("() !a")
            not_neg = "(Next (Not Proposition(a)))"
            self.assertEqual(not_neg, str(formula.nnf(False)))
            
            negated = "(Next Proposition(a))"
            self.assertEqual(negated, str(formula.nnf(True)))
    
    def verify_step_fairness_constraint(self):
        # must be true
        tool = ast.fairness_constraint(self.befsm, 0, 0)
        self.assertEqual(tool, Be.true(self.mgr))
         
        # loop position does not matter if not feasible
        tool = ast.fairness_constraint(self.befsm, 0, 1)
        self.assertEqual(tool, Be.true(self.mgr))
        
        model= bmcutils.BmcModel()
        # step 0
        tool = ast.fairness_constraint(self.befsm, 1, 0)
        smv  = model.fairness(1, 0) 
        self.assertEqual(tool, smv)
         
        # step 1
        tool = ast.fairness_constraint(self.befsm, 2, 1)
        smv  = model.fairness(2, 1) 
        self.assertEqual(tool, smv)
        
    def test_step_fairness_constraint(self):
        with tests.Configure(self, __file__, "/example.smv"):
            self.verify_step_fairness_constraint()
        
        with tests.Configure(self, __file__, "/philo.smv"):
            self.verify_step_fairness_constraint()


import cmd
import re
from .lockStack import LockStack
from ..ast import Spec
from ..lazyTlace.explain import explain_branch
from ..lazyTlace.tlace import (PartialTlacenode, Tlacebranch,
                               TemporalBranch, EpistemicBranch)

class LazyCTLK_explain_shell(cmd.Cmd):
    """
    A shell for CTLK TLACE explanation.
    """
    
    def __init__(self, fsm, tlace):
        super(LazyCTLK_explain_shell, self).__init__()
        self.prompt = ">> "
        self.tlace = tlace
        self.stack = LockStack()
        self.stack.push(tlace)
        self.fsm = fsm
        
        
    def do_help(self, arg):
        """
        Display help for this shell.
        
        arg argument is ignored.
        """
        
        # Explain the syntax for explanations
        
        # Syntax follows this regex: (\.|[0-9]+)*
        
        # Semantics: show the corresponding object
        # a number selects the corresponding branch/node
        # a dot separates two numbers
        # more dots go up the tree
        
        # reset, empty line, CTRL-d
        
        help = """This prompt allows to explore the TLACE explaining a property.
It starts at the root of the TLACE and allows navigation.

The commands to navigate follow this regular expression (\.|[0-9]+)*
    a dot selects the current node/branch, more dots go up to the parent;
    a number selects the branch/node of the current item;
    
    ex: 0.3.0.1 from the root node selects the second node of the first branch
        of the third node of the first branch of the root node;
        ...1 from a node selects the second branch of the parent of its parent
        branch;
        
In addition, three commands are also provided
    reset: restart the explanation to the root node;
    an empty line: shows the current item;
    CTRL-d: quits the prompt to return to the CLI main prompt.
"""
        print(help)
        
        
    def do_reset(self, arg):
        """Reset the explanation to top TLACE node."""
        self.stack = LockStack()
        self.stack.push(self.tlace)
        self._show(self.stack.top())
        
        
    def emptyline(self):
        """Show current node/branch."""
        self._show(self.stack.top())
        return False
        
    
    def default(self, line):
        """Interpret line as a request for explanation."""
        
        # Pre-work
        self.stack.unlock()
        stack = self.stack.copy() # Copy the stack to restore it if error
        
        # Check syntax: (\.|[0-9]+)*
        if not re.match("^(\.|[0-9]+)*$", line):
            print("error: wrong format")
            return False
        
        # Check existence of requested node and branch
        try:
            elements = []
            for (grp, i) in zip(re.finditer(r"\.|[0-9]+", line),
                                  range(len(line))):
                index = grp.group(0)
                elements.append(index)
                if index == ".":
                    stack.pop()
                else:
                    # Check existence and push corresponding element
                    index = int(index)
                    elem = stack.top()
                    if isinstance(elem, PartialTlacenode):
                        if index < 0 or len(elem.branches) <= index:
                            print("error: index out of range:",
                                  "".join(elements))
                            return False
                        else:
                            if isinstance(elem.branches[index], Spec):
                                elem.branches[index] = explain_branch(
                                        self.fsm, elem.state,
                                        elem.branches[index],
                                        elem.branches[index])
                            stack.push(elem.branches[index])
                    elif isinstance(elem, Tlacebranch):
                        index = index * 2
                        if index < 0 or len(elem.path) <= index:
                            print("error: index out of range",
                                  "".join(elements))
                            return False
                        else:
                            stack.push(elem.path[index])
            
        except IndexError:
            print("error: cannot find parent of root TLACE node")
            return False
        
        # Explain the branch
        try:
            elem = stack.top()
            if type(elem) is PartialTlacenode:
                print("Explaining node")
            elif type(elem) is TemporalBranch:
                print("Explaining temporal branch for", elem.specification)
            elif type(elem) is EpistemicBranch:
                print("Explaining epistemic branch for", elem.specification)
            print("-" * 80)
            self._show(elem)
        except IndexError:
            print("error: reached root TLACE node and beyond; abort")
            return False
            
        # Save the modified stack
        self.stack = stack
        
        
    def do_EOF(self, arg):
        """Quit the explanation prompt (you can hit CTRL-d)."""
        print()
        return True
            
    
    def _show(self, tlace, index=0, prev=None, prefix=None):
        """Show the given TLACE element, a TLACE node or branch."""
        prefix = prefix if prefix else ""
        if type(tlace) is PartialTlacenode:
            # State
            header = " State " + str(index) + " "
            print(header.center(40,"-"))
            values = tlace.state.get_str_values()
            for var in values:
                if (prev is not None and
                    prev.state.get_str_values()[var] != values[var] or
                    prev is None) :
                    print(var, "=", values[var])
            
            # Annotations
            if len(tlace.atomics) + len(tlace.branches) + len(tlace.universals):
                header = "--- Annotations"
                print(header)
                for atomic in tlace.atomics:
                    print(atomic)
                for (branch, i) in zip(tlace.branches,
                                                    range(len(tlace.branches))):
                    if isinstance(branch, Spec):
                        print(prefix + str(i),":", branch)
                    elif isinstance(branch, Tlacebranch):
                        print(prefix + str(i),":", branch.specification)
                for univ in tlace.universals:
                    print(univ)
                    
        elif type(tlace) is TemporalBranch:
            # Loop?
            if tlace.loop is not None and tlace.loop[1] == tlace.path[0]:
                header = " Loop starts here "
                print(header.center(40, "-"))
            self._show(tlace.path[0], prefix=prefix + "0.")
            # Show the states, and the inputs,
            for (os, i, s, ind) in zip(tlace.path[::2], tlace.path[1::2],
                                       tlace.path[2::2],
                                       range(len(tlace.path[2::2]))):
                # Inputs
                header = " Inputs "
                print(header.center(40, "-"))
                values = i.get_str_values()
                for var in values:
                    print(var, "=", values[var])
                # don't forget loops,
                if tlace.loop is not None and tlace.loop[1] == s:
                    header = " Loop starts here "
                    print(header.center(40, "-"))
                self._show(s, ind + 1, os, prefix + str(ind + 1) + ".")
            # Print loop inputs if needed
            if tlace.loop is not None:
                header = " Inputs "
                print(header.center(40, "-"))
                values = tlace.loop[0].get_str_values()
                for var in values:
                    print(var, "=", values[var])
                
        elif type(tlace) is EpistemicBranch:
            # Show the states, and the agents,
            self._show(tlace.path[0])
            for (os, ag, s, ind) in zip(tlace.path[::2], tlace.path[1::2],
                                        tlace.path[2::2],
                                        range(len(tlace.path[2::2]))):
                # Agents
                if len(ag) > 1:
                    header = " Agents : "
                else:
                    header = " Agent : "
                header += ",".join(ag) + " "
                print(header.center(40, "-"))
                # State
                self._show(s, ind + 1, os, prefix + str(ind + 1) + ".")


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_lower_interface.nusmv.dd import dd as nsdd

from pynusmv_tools.mas import glob

from pynusmv_tools.atlkPO.eval import split
from pynusmv_tools.atlkPO.evalPartial import split as split_partial


class TestSplit(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
    
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
    
    
    def little(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/little.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame3(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame3.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        for si in fsm.pick_all_states_inputs(bdd):
            print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        for s in fsm.pick_all_states(bdd):
            print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        for i in fsm.pick_all_inputs(bdd):
            print(i.get_str_values())
    
    @unittest.skip    
    def test_split_by_picking(self):
        fsm = self.little()
        
        aa = eval_simple_expression(fsm, "a.a = 1")
        ap = eval_simple_expression(fsm, "a.p = 1")
        ba = eval_simple_expression(fsm, "b.a = 1")
        bq = eval_simple_expression(fsm, "b.q = 1")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        excluded = BDD.false(fsm.bddEnc.DDmanager)
        
        strats = fsm.protocol({"a"})
        print("protocol of a -- strats to consider")
        self.show_si(fsm, strats)
        si = fsm.pick_one_state_inputs(strats)
        self.assertEqual(len(fsm.pick_all_states(
                                            si.forsome(fsm.bddEnc.inputsCube))),
                         1)
        s = fsm.pick_one_state(si.forsome(fsm.bddEnc.inputsCube))
        print("si:", si.get_str_values())
        print("s:", s.get_str_values())
        eqs = fsm.equivalent_states(s, {"a"})
        print("equivalent to", s.get_str_values())
        self.show_s(fsm, eqs)
        eqcl = strats & eqs
        print("equivalence class (" + str(s.get_str_values()) + ")")
        self.show_si(fsm, eqcl)
        act = eqcl.forsome(fsm.bddEnc.statesCube)
        print("actions of eq class (" + str(fsm.count_inputs(act)) + ")")
        self.show_i(fsm, act)
        
        
        gamma_inputs = [var
                        for agent in {"a"}
                        for var in fsm.agents_inputvars[agent]]
        gamma_cube = fsm.bddEnc.cube_for_inputs_vars(gamma_inputs)
        ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
        
        
        nact = ~act.forsome(ngamma_cube) # Useless here
        print("the other actions")
        self.show_i(fsm, nact)
        
        # There is more than one action if act contains other actions than
        # si
        self.assertTrue((act -
                         si.forsome(fsm.bddEnc.statesCube).
                            forsome(ngamma_cube)).is_false())
        print("Are there other actions than si's one ?",
              (act - si.forsome(fsm.bddEnc.statesCube).
                                  forsome(ngamma_cube)).isnot_false())
        
        # remove act from Strats
        excluded = excluded | (strats & eqcl)
        print("excluded")
        self.show_si(fsm, excluded)
        strats = strats - excluded
        print("new strats")
        self.show_si(fsm, strats)
        
        # Restart the process
        si = fsm.pick_one_state_inputs(strats)
        self.assertEqual(len(fsm.pick_all_states(
                                            si.forsome(fsm.bddEnc.inputsCube))),
                         1)
        s = fsm.pick_one_state(si.forsome(fsm.bddEnc.inputsCube))
        print("si:", si.get_str_values())
        print("s:", s.get_str_values())
        eqs = fsm.equivalent_states(s, {"a"})
        print("equivalent to", s.get_str_values())
        self.show_s(fsm, eqs)
        eqcl = strats & eqs
        print("equivalence class (" + str(s.get_str_values()) + ")")
        self.show_si(fsm, eqcl)
        act = eqcl.forsome(fsm.bddEnc.statesCube)
        print("actions of eq class (" + str(fsm.count_inputs(act)) + ")")
        self.show_i(fsm, act)
        
        print("Are there other actions than si's one ?",
              (act - si.forsome(fsm.bddEnc.statesCube).
                                  forsome(ngamma_cube)).isnot_false())
        self.show_i(fsm, (act - si.forsome(fsm.bddEnc.statesCube).
                                   forsome(ngamma_cube)))
                                   
        # Need to split eqcl now !
        # Keep eqcl to split strats later
        fulleqcl = eqcl
        ncss = (eqcl & si.forsome(fsm.bddEnc.statesCube).
                          forsome(ngamma_cube))
        eqcls = [ncss]
        eqcl = eqcl - ncss
        while eqcl.isnot_false():
            si = fsm.pick_one_state_inputs(eqcl)
            ncss = (eqcl & si.forsome(fsm.bddEnc.statesCube).
                              forsome(ngamma_cube))
            eqcls.append(ncss)
            eqcl = eqcl - ncss
        for ncss in eqcls:
            print("new non-conflicting subset")
            self.show_si(fsm, ncss)
            
        splitted = []
        for ncss in eqcls:
            splitted.append(strats - fulleqcl + ncss + excluded)
        print("show splitted strats")
        for substrat in splitted:
            print("new sub strategy")
            self.show_si(fsm, substrat)
            
            
    def test_split(self):
        fsm = self.little()
        
        aa = eval_simple_expression(fsm, "a.a = 1")
        ap = eval_simple_expression(fsm, "a.p = 1")
        ba = eval_simple_expression(fsm, "b.a = 1")
        bq = eval_simple_expression(fsm, "b.q = 1")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        gamma_inputs = [var
                        for agent in {"a"}
                        for var in fsm.agents_inputvars[agent]]
        gamma_cube = fsm.bddEnc.cube_for_inputs_vars(gamma_inputs)
        ngamma_cube = fsm.bddEnc.inputsCube - gamma_cube
        
        strats = split(fsm, fsm.protocol({"a"}), {"a"})
            
        commstrat = (((~ap & bq & ~aa) | (~ap & ~bq & ~aa)).forsome(ngamma_cube) &
                     fsm.protocol({"a"}))
        firststrat = (((ap & bq & aa) | (ap & ~bq & aa)).forsome(ngamma_cube) &
                      fsm.protocol({"a"}))
        secstrat = (((ap & bq & ~aa) | (ap & ~bq & ~aa)).forsome(ngamma_cube) &
                    fsm.protocol({"a"}))
        
        self.assertTrue((commstrat | firststrat) in strats)
        self.assertTrue((commstrat | secstrat) in strats)
        
        self.assertSetEqual({commstrat | firststrat, commstrat | secstrat},
                            strats)
    
    def test_split_cardgame3(self):
        fsm = self.cardgame3()
        agents = {'player'}
        
        strats = set()
        nbstrats = 0
        for strat in split(fsm, fsm.protocol(agents), agents):
            nbstrats += 1
            self.assertTrue(strat not in strats)
            strats.add(strat)
        self.assertEqual(nbstrats, 8)
        self.assertEqual(len(strats), nbstrats)
        


import unittest

from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_tools.mas import glob
from pynusmv_tools.ctlk.eval import evalCTLK
from pynusmv_tools.ctlk.parsing import parseCTLK

from pynusmv_tools.ctlk.eval import nd

class TestEval(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        
    def tearDown(self):
        glob.reset_globals()
        deinit_nusmv()
        
    def model(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/dining-crypto.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def simplemodel(self):
        glob.load_from_file("tests/pynusmv_tools/ctlk/agents.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
        
    def test_true(self):
        fsm = self.model()
        
        true = eval_simple_expression(fsm, "TRUE")
        
        specs = parseCTLK("True")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        self.assertEqual(true, evalCTLK(fsm, spec))
        
        
    def test_false(self):
        fsm = self.model()
        
        false = eval_simple_expression(fsm, "FALSE")
        
        specs = parseCTLK("False")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        self.assertEqual(false, evalCTLK(fsm, spec))
        
        
    def test_init(self):
        fsm = self.model()
        
        specs = parseCTLK("Init")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        self.assertEqual(fsm.init, evalCTLK(fsm, spec))
        
        
    def test_reachable(self):
        fsm = self.model()
        
        specs = parseCTLK("Reachable")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        self.assertEqual(fsm.reachable_states, evalCTLK(fsm, spec))
        
    
    def test_atom(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        
        specs = parseCTLK("'c1.payer'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        c1pCTLK = evalCTLK(fsm, spec)
        self.assertEqual(c1p, c1pCTLK)
        
        
    def test_not(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        
        specs = parseCTLK("~'c1.payer'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        c1pCTLK = evalCTLK(fsm, spec)
        self.assertEqual(~c1p, c1pCTLK)
        
        
    def test_and(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        odd = eval_simple_expression(fsm, "countsay = odd")
        
        specs = parseCTLK("'c1.payer' & 'countsay = odd'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        c1paco = evalCTLK(fsm, spec)
        self.assertEqual(c1p & odd, c1paco)
        
               
    def test_or(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        odd = eval_simple_expression(fsm, "countsay = odd")
        
        specs = parseCTLK("'c1.payer' | 'countsay = odd'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        c1poco = evalCTLK(fsm, spec)
        self.assertEqual(c1p | odd, c1poco)
        
    
    def test_implies(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        odd = eval_simple_expression(fsm, "countsay = odd")
        
        specs = parseCTLK("'c1.payer' -> 'countsay = odd'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        c1pico = evalCTLK(fsm, spec)
        self.assertEqual(c1p.imply(odd), c1pico)
        
        
    def test_iff(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        odd = eval_simple_expression(fsm, "countsay = odd")
        
        specs = parseCTLK("'c1.payer' <-> 'countsay = odd'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        
        c1pico = evalCTLK(fsm, spec)
        self.assertEqual(c1p.iff(odd), c1pico)
        
        
    def test_ex(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        
        specs = parseCTLK("EX ('c1.payer' & ~'c2.payer' & ~'c3.payer')")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        ex = evalCTLK(fsm, spec)
        self.assertEqual(ex, c1p & ~c2p & ~c3p & fsm.bddEnc.statesMask)
        
        specs = parseCTLK("EX 'countsay = odd'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        ex = evalCTLK(fsm, spec)
            
        self.assertTrue(odd & fsm.bddEnc.statesMask <= ex)
        self.assertTrue((unk & c1p & ~c2p & ~c3p) & fsm.bddEnc.statesMask
                        <= ex)
        self.assertTrue((unk & c1p & c2p & c3p) & fsm.bddEnc.statesMask
                        <= ex)
        
        
    def test_simple_ex(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        
        specs = parseCTLK("EX 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        ex = evalCTLK(fsm, spec)
        # ex : 
        #  l1, !l2, !g
        #  !l1, l2, g
        #  l1, l2, g
        #  !l1, !l2, !g
        self.assertEqual(ex, lf.iff(g))
        
        specs = parseCTLK("EX ('af.local' & 'at.local' & 'global')")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        ex = evalCTLK(fsm, spec)
        self.assertEqual(ex, lt & ~lf & ~g | ~lt & lf & g)
        
        
    def test_k_ex(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        
        specs = parseCTLK("K<'at'> EX 'global'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        kex = evalCTLK(fsm, spec)
        self.assertEqual(kex, true)
        
        specs = parseCTLK("K<'at'> EX ~'global'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        kexn = evalCTLK(fsm, spec)
        self.assertEqual(kexn, true)
        
        
    def test_k_ef(self):
        fsm = self.simplemodel()
        
        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        
        specs = parseCTLK("K<'at'> EF 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        kef = evalCTLK(fsm, spec)
        self.assertEqual(kef, true)
        
        
    def test_k(self):
        fsm = self.simplemodel()

        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")

        specs = parseCTLK("K<'at'> 'at.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        katt = evalCTLK(fsm, spec)
        self.assertEqual(katt, lt)
        
        specs = parseCTLK("K<'at'> 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        katf = evalCTLK(fsm, spec)
        self.assertEqual(katf, false)
        
        specs = parseCTLK("K<'at'> 'global'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        katf = evalCTLK(fsm, spec)
        self.assertEqual(katf, g)
        
        
    def test_k_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        specs = parseCTLK("AG ('c1.payer' "
                                      "-> K<'c1'> (~'c2.payer' & ~'c3.payer'))")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        ik = evalCTLK(fsm, spec)                
        self.assertTrue(fsm.init <= ik)
        
    
    def test_e_simple(self):
        fsm = self.simplemodel()

        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")

        specs = parseCTLK("'global' -> E<'at','af'> 'global'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        geg = evalCTLK(fsm, spec) 
        self.assertEqual(geg, true)
        
        specs = parseCTLK("'af.local' & E<'at','af'> 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        afeaf = evalCTLK(fsm, spec)
        self.assertEqual(afeaf, false)
        
        
    def test_e_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        specs = parseCTLK("'countsay = even' -> E<'c1','c2'> ~'c1.payer'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        een1 = evalCTLK(fsm, spec)                
        self.assertEqual(een1, true)
        
        specs = parseCTLK("E<'c2','c3'> ~'c1.payer'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        en1 = evalCTLK(fsm, spec)  
        self.assertTrue(even & fsm.reachable_states <= en1)
        
        
    def test_d_simple(self):
        fsm = self.simplemodel()

        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")

        specs = parseCTLK("'af.local' -> D<'at','af'> 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        geg = evalCTLK(fsm, spec) 
        self.assertEqual(geg, true)
        
        specs = parseCTLK("'af.local' & D<'at'> 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        geg = evalCTLK(fsm, spec) 
        self.assertEqual(geg, false)
        
        specs = parseCTLK("'af.local' -> D<'af'> 'af.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        geg = evalCTLK(fsm, spec) 
        self.assertEqual(geg, true)
        
        
    def test_dincry_equiv(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        s = c1p & ~c2p & ~c3p
        eqs2 = fsm.equivalent_states(s, {"c2"})
        eqs3 = fsm.equivalent_states(s, {"c3"})
        self.assertEqual(eqs2 & eqs3, ~c2p & ~c3p)
    
    
    def test_d_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        specs = parseCTLK("('c1.payer' & 'countsay != unknown') "
                          "-> D<'c2','c3'> 'c1.payer'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        cud1 = evalCTLK(fsm, spec)        
        self.assertTrue(fsm.reachable_states <= cud1)


    def test_d_distributedmodel(self):  
        glob.load_from_file("tests/pynusmv_tools/ctlk/distributed_knowledge.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        
        l1 = eval_simple_expression(fsm, "a1.local")
        l2 = eval_simple_expression(fsm, "a2.local")
        l3 = eval_simple_expression(fsm, "a3.local")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        specs = parseCTLK("'a1.local' -> D<'a1','a2'> 'a1.local'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        a1da1 = evalCTLK(fsm, spec)        
        self.assertEqual(true, a1da1)
                
        
    def test_c_simple(self):
        fsm = self.simplemodel()

        lt = eval_simple_expression(fsm, "at.local")
        lf = eval_simple_expression(fsm, "af.local")
        g = eval_simple_expression(fsm, "global")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")

        specs = parseCTLK("'global' -> C<'at','af'> 'global'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        gcg = evalCTLK(fsm, spec) 
        self.assertEqual(gcg, true)
        
        specs = parseCTLK("C<'at','af'> 'global' | C<'at','af'> ~'global'")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        gcg = evalCTLK(fsm, spec) 
        self.assertEqual(gcg, true)
        
        
    def test_c_dincry(self):
        fsm = self.model()
        
        c1p = eval_simple_expression(fsm, "c1.payer")
        c2p = eval_simple_expression(fsm, "c2.payer")
        c3p = eval_simple_expression(fsm, "c3.payer")
        c1h = eval_simple_expression(fsm, "c1.coin = head")
        c2h = eval_simple_expression(fsm, "c2.coin = head")
        c3h = eval_simple_expression(fsm, "c3.coin = head")
        odd = eval_simple_expression(fsm, "countsay = odd")
        even = eval_simple_expression(fsm, "countsay = even")
        unk = eval_simple_expression(fsm, "countsay = unknown")
        true = eval_simple_expression(fsm, "TRUE")
        false = eval_simple_expression(fsm, "FALSE")
        
        specs = parseCTLK("'countsay = odd' -> "
                          "C<'c1','c2','c3'> "
                          "('c1.payer' | 'c2.payer' | 'c3.payer')")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        oc123 = evalCTLK(fsm, spec)        
        self.assertTrue(fsm.reachable_states <= oc123)
        
        specs = parseCTLK("'countsay = even' -> "
                          "C<'c1','c2','c3'> "
                          "(~'c1.payer' & ~'c2.payer' & ~'c3.payer')")
        self.assertEqual(len(specs), 1)
        spec = specs[0]
        ec123 = evalCTLK(fsm, spec)        
        self.assertTrue(fsm.reachable_states <= ec123)


import unittest

from pynusmv.dd import BDD
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv.mc import eval_simple_expression

from pynusmv_lower_interface.nusmv.dd import dd as nsdd

from pynusmv_tools.mas import glob

from pynusmv_tools.atlkPO.check import check
from pynusmv_tools.atlkFO.parsing import parseATLK

from pynusmv_tools.atlkPO import config


class TestCheckPartial(unittest.TestCase):
    
    def setUp(self):
        init_nusmv()
        #config.debug = True
        config.partial.early.type = "full"
        config.partial.caching = True
        config.partial.filtering = True
        #config.partial.separation.type = "reach"
        #config.garbage.type = "step"
        #config.garbage.step = 4
        config.partial.alternate.type = {"univ", "strat"}
    
    def tearDown(self):    
        glob.reset_globals()
        deinit_nusmv()
    
    
    def little(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/little.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def cardgame_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/cardgame-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def trans2_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/2-transmission-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def transmission_with_knowledge(self):
        glob.load_from_file(
                        "tests/pynusmv_tools/atlkPO/models/transmission-knowledge.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
    
    def transmission_post_fair(self):
        glob.load_from_file("tests/pynusmv_tools/atlkPO/models/transmission-post-fair.smv")
        fsm = glob.mas()
        self.assertIsNotNone(fsm)
        return fsm
        
    def show_si(self, fsm, bdd):
        if bdd.isnot_false():
            sis = fsm.pick_all_states_inputs(bdd)
            print("SI count:", len(sis))
            for si in sis:
                print(si.get_str_values())
            
    def show_s(self, fsm, bdd):
        if bdd.isnot_false():
            for s in fsm.pick_all_states(bdd):
                print(s.get_str_values())
            
    def show_i(self, fsm, bdd):
        if bdd.isnot_false():
            for i in fsm.pick_all_inputs(bdd):
                print(i.get_str_values())
        
        
    def test_cardgame_not_improved_partial(self):
        fsm = self.cardgame()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial"))
        
        self.assertFalse(check(fsm, parseATLK("AG <'player'> X 'pcard = Ac'")[0], implem="partial"))
        
    
    def test_cardgame_post_fair_not_improved_partial(self):
        fsm = self.cardgame_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial"))
        

    def test_transmission_not_improved_partial(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial"))
        
    
    def test_transmission_with_know_not_improved_partial(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial"))
    
    
    def test_transmission_post_fair_not_improved_partial(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial"))
        
        
    def test_tpfni_partial_vs_normal(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0]))
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial"))
    
    
    def test_cardgame_improved_partial(self):
        fsm = self.cardgame()
        
        self.assertFalse(check(fsm, parseATLK("<'player'> X 'win'")[0],
                         implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("AG <'player'> X 'pcard = Ac'")[0], implem="partial", variant="FS"))
        
    
    def test_cardgame_post_fair_improved_partial(self):
        fsm = self.cardgame_post_fair()
        
        self.assertFalse(check(fsm, parseATLK("<'player'> X 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial", variant="FS"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial", variant="FS"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial", variant="FS"))
        

    def test_transmission_improved_partial(self):
        fsm = self.transmission()
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial", variant="FS"))
        
    
    def test_transmission_with_know_improved_partial(self):
        fsm = self.transmission_with_knowledge()
        
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
    
    
    def test_transmission_post_fair_improved_partial(self):
        fsm = self.transmission_post_fair()
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial", variant="FS"))
        
    
    def test_cardgame_improved_partial_univ_only(self):
        fsm = self.cardgame()
        config.partial.alternate.type = {"univ"}
        
        self.assertFalse(check(fsm, parseATLK("<'player'> X 'win'")[0],
                         implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("AG <'player'> X 'pcard = Ac'")[0], implem="partial", variant="FS"))
        
    
    def test_cardgame_post_fair_improved_partial_univ_only(self):
        fsm = self.cardgame_post_fair()
        config.partial.alternate.type = {"univ"}
        
        self.assertFalse(check(fsm, parseATLK("<'player'> X 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial", variant="FS"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial", variant="FS"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial", variant="FS"))
        

    def test_transmission_improved_partial_univ_only(self):
        fsm = self.transmission()
        config.partial.alternate.type = {"univ"}
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial", variant="FS"))
        
    
    def test_transmission_with_know_improved_partial_univ_only(self):
        fsm = self.transmission_with_knowledge()
        config.partial.alternate.type = {"univ"}
        
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
    
    
    def test_transmission_post_fair_improved_partial_univ_only(self):
        fsm = self.transmission_post_fair()
        config.partial.alternate.type = {"univ"}
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial", variant="FS"))
    
    
    def test_cardgame_improved_partial_strat_only(self):
        fsm = self.cardgame()
        config.partial.alternate.type = {"strat"}
        
        self.assertFalse(check(fsm, parseATLK("<'player'> X 'win'")[0],
                         implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial", variant="FS"))
        
        self.assertFalse(check(fsm, parseATLK("AG <'player'> X 'pcard = Ac'")[0], implem="partial", variant="FS"))
        
    
    def test_cardgame_post_fair_improved_partial_strat_only(self):
        fsm = self.cardgame_post_fair()
        config.partial.alternate.type = {"strat"}
        
        self.assertFalse(check(fsm, parseATLK("<'player'> X 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("K<'player'>'pcard=none' & K<'player'>'dcard=none'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~(K<'player'> 'dcard=Ac' | K<'player'> 'dcard=K' | K<'player'> 'dcard=Q'))")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("AG('step = 1' -> ~<'player'> X 'win')")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("['dealer'] F 'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("['player'] X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> X 'pcard=Ac'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'dealer'> G ~'win'")[0], implem="partial", variant="FS"))
        
        # Player can win
        self.assertTrue(check(fsm, parseATLK("<'player'> F 'win'")[0], implem="partial", variant="FS"))
        # Dealer can avoid fairness
        self.assertTrue(check(fsm, parseATLK("<'dealer'> F 'FALSE'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EG ~'win'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EF 'win'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> G <'player'> F 'win'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("<'player'> F <'player'>[~'lose' U 'win']")[0], implem="partial", variant="FS"))
        

    def test_transmission_improved_partial_strat_only(self):
        fsm = self.transmission()
        config.partial.alternate.type = {"strat"}
        
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        # False because transmitter cannot win if received is already true
        # and he has no clue about it
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial", variant="FS"))
        
    
    def test_transmission_with_know_improved_partial_strat_only(self):
        fsm = self.transmission_with_knowledge()
        config.partial.alternate.type = {"strat"}
        
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
    
    
    def test_transmission_post_fair_improved_partial_strat_only(self):
        fsm = self.transmission_post_fair()
        config.partial.alternate.type = {"strat"}
        
        self.assertTrue(check(fsm, parseATLK("<'sender'> F 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> G ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("<'sender'> X 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> X ~'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("<'transmitter'> F 'received'")[0], implem="partial", variant="FS"))
        
        # False because the sender does not know if the bit is already
        # transmitted (and in this case, no strategy can avoid 'received')
        self.assertFalse(check(fsm, parseATLK("<'sender'> G ~'received'")[0], implem="partial", variant="FS"))
        
        self.assertTrue(check(fsm, parseATLK("EF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("EG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AG ~'received'")[0], implem="partial", variant="FS"))
        self.assertFalse(check(fsm, parseATLK("AF 'received'")[0], implem="partial", variant="FS"))
        self.assertTrue(check(fsm, parseATLK("AF ~'received'")[0], implem="partial", variant="FS"))


"""
This module contains the class definition of the different classes that may
compose an LTL abstract syntax tree.

.. note::
    The generation of the boolean expressions corresponding to the different
    LTL expressions could have been done using directly the functionalities
    provided :mod:`pynusmv.bmc.ltlspec`. However, the objective of the tool
    implemented in :mod:`pynusmv_tools.bmcLTL` is to demonstrate with a simple example
    (LTL) how one can use the bmc additions to PyNuSMV to develop verification
    tools for new logic formalisms. Hence, the approach that was adopted in that
    context was to *NOT* use any of the "higher level services" and develop
    a verification tool translating "literally" the reduction of that logic to
    a propositional SAT problem as imagined or stated in a reference paper.

    In this scope, the reference paper that was used was:

                Biere et al - ``Bounded Model Checking'' - 2003
"""
from pynusmv.bmc.glob      import master_be_fsm
from pynusmv.be.expression import Be
from pynusmv.wff           import Wff
from pynusmv.parser import parse_simple_expression

###############################################################################
# Memoization Utility functions:
###############################################################################
# The global cache of memoized values
MEMOIZER = {}

# creates a key for a cache entry
def MEMOIZER_key(fml, i, k, l):
    """
    Creates a key for a cache entry.
    
    :param fml: the formula object (typically 'self').
    :param i: the time index.
    :param k: the problem bound.
    :param l: the loop position. Set this value to -1 in the stright path case.
    :return: a key for the memoized entry _l[[fml]]^i_k
    """
    return (fml, i, k, l)

def MEMOIZER_get(key):
    """
    Retrieves a memoized entry by its key.
    
    :param key: the key (created with MEMOIZER_key) of the value to retrieve
    :return: the memoized value corresponding to key or none if nothing was found
    """
    return MEMOIZER.get(key)

def MEMOIZE_put(key, value):
    """
    Memoises `value` and associates it with `key`.
    
    :param key: the key (created with MEMOIZER_key) of the value to store
    :param value: the value to store
    """
    MEMOIZER[key] = value

def memoize_with_loop(fun):
    """
    The decorator to be applied on semantic_with_loop functions to activate the
    memoisation (makes the reduction to SAT linear).
    
    :param fun: the decorated function. (should be of the form:
        fun(fml, enc, i, k, l)
    """
    def memoized(fml, enc, i, k, l):
        key = MEMOIZER_key(fml, i, k, l)
        res = MEMOIZER_get(key)
        if res is None:
            res = fun(fml, enc, i, k, l)
            MEMOIZE_put(key, res)
        return res
    return memoized

def memoize_no_loop(fun):
    """
    The decorator to be applied on semantic_no_loop functions to activate the
    memoisation (makes the reduction to SAT linear).
    
    :param fun: the decorated function. (should be of the form:
        fun(fml, enc, i, k)
    """
    def memoized(fml, enc, i, k):
        key = MEMOIZER_key(fml, i, k, -1)
        res = MEMOIZER_get(key)
        if res is None:
            res = fun(fml, enc, i, k)
            MEMOIZE_put(key, res)
        return res
    return memoized   

###############################################################################
# LTL Utility functions:
###############################################################################
def successor(i, k, l):
    """
    Computes the successor of time `i` in a k-l loop.

    .. note::
        References, see Definition 6
        in Biere et al - ``Bounded Model Checking'' - 2003

    .. note::
        An other implementation of this function (w/ the same semantics) exists
        in :mod:`pynusmv.bmc.utils`. This version is merely re-implemented to
        1. show that it can be easily done
        2. stick closely to the definition given in the paper by Biere et al.
            (see other note)

    .. warning::
        To be consistent with the way the loop condition is implemented (equiv
        of all the state variables), we have that walking 'k' steps means to be
        back at step 'l'. Hence, the value of i can only vary from 0 to k-1
        (and will repeat itself in the range [l; k-1]

    :param i: the current time
    :param k: the maximum (horizon/bound) time of the problem
    :param l: the time where the loop starts

    :return: the k-l loop successor of i
    """
    return i+1 if i < k-1 else l

def loop_condition(enc, k, l):
    """
    This function generates a Be expression representing the loop condition
    which is necessary to determine that k->l is a backloop.

    Formally, the returned constraint is denoted _{l}L_{k}

    Because the transition relation is encoded in Nusmv as formula (and not as
    a relation per-se), we determine the existence of a backloop between
    l < k and forall var, var(i) == var(k)

    That is to say: if it is possible to encounter two times the same state
    (same state being all variables have the same value in both states) we know
    there is a backloop on the path

    .. note::
        An other implementation of this function (w/ the same semantics) exists
        in :mod:`pynusmv.bmc.utils`. This version is merely re-implemented to
        1. show that it can be easily done
        2. stick closely to the definition given in the paper by Biere et al.
            (see other note)


    :param fsm: the fsm on which the condition will be evaluated
    :param k: the highest time
    :param l: the time where the loop is assumed to start
    :return: a Be expression representing the loop condition that verifies that
        k-l is a loop path.
    """
    cond = Be.true(enc.manager)
    for v in enc.curr_variables: # for all untimed variable
        vl   = v.at_time[l].boolean_expression
        vk   = v.at_time[k].boolean_expression
        cond = cond & ( vl.iff(vk) )
    return cond

def fairness_constraint(fsm, k, l):
    """
    Computes a step of the constraint to be added to the loop side of the BE
    when one wants to take fairness into account for the case where we consider
    the existence of a k-l loop (between k and l obviously).

    :param fsm: the fsm whose transition relation must be unrolled
    :param k: the maximum (horizon/bound) time of the problem
    :param l: the time where the loop starts
    :return: a step of the fairness constraint to force fair execution on the
        k-l loop.
    """
    constraint = Be.true(fsm.encoding.manager)
    # nothing to generate, stop
    if k == 0:
        return constraint

    for fairness in fsm.fairness_iterator():
        # just a shortcut for the loop to create
        #    \bigvee_{l}^{k-1} (fairness_{l})
        constraint &= fsm.encoding.or_interval(fairness, l, k-1)
    return constraint

###############################################################################
# Abstract ast nodes
###############################################################################
class Formula:
    """An abstract base class meant to be the parent of all the AST nodes"""

    def bounded_semantics(self, fsm, k, fairness=True):
        """
        Returns a boolean expression corresponding to the bounded semantics of
        the formula denoted by `self` on a path of length k. This combines both
        the semantics in case of a loopy path and the case of a non-loopy path.

        .. note::
            This function takes the same approach as NuSMV and does not enforce
            the absence of loop when using the more restrictive semantic_no_loop.

        :param fsm: the FSM representing the model. It is used to gain access
            to the encoder (-> shift variables) and to obtain the list of
            fairness constraints.
        :param k: the last time that exists in the universe of this expression
        :param fairness: a flag indicating whether or not the fairness constraints
            should be taken into account while generating the formula.
        :return: a boolean expression translating the bounded semantics of this
            formula.
        """
        enc    = fsm.encoding
        noloop = self.semantic_no_loop(enc, 0, k)

        w_loop = Be.false(enc.manager)
        for l in range(k):   # [0; k-1]
            fairness_cond = fairness_constraint(fsm, k, l) \
                                 if fairness \
                                 else Be.true(enc.manager)

            w_loop |= (loop_condition(enc, k, l) \
                      & fairness_cond \
                      & self.semantic_with_loop(enc, 0, k, l))

        return noloop | w_loop

    def semantic_no_loop(self, enc, i, k):
        """
        All nodes of the AST must implement this function.
        Concretely, the role of this function is to generate a propositional
        equivalent to the bounded LTL semantic of this node when there is no
        loop on the path from state(time) to state(bound)

        ..math::
            [[self]]_{k}^{i}

        :param enc: the encoding used to store and organize the variables (used
            ie to shift vars)
        :param i: the time at which the generated expression will be considered
        :param k: the last time that exists in the universe of this expression
        :return: a boolean expression conform to the ltl bounded semantics of
            this node when there is no loop on the path from i to k
        """
        pass

    def semantic_with_loop(self, enc, i, k, l):
        """
        All nodes of the AST must implement this function.
        Concretely, the role of this function is to generate a propositional
        equivalent to the bounded LTL semantic of this node when there is no
        loop on the path from state(time) to state(bound)

        ..math::
            _{l}[[self]]_{k}^{i}

        .. note::
            There are essentially two cases for the loop:

                1. i < l         (not yet entered in the loop)
                2. l <= i < k    (inside the loop)

            Therefore, l is allowed to range from 0 to k. However, you don't
            need to deal with all the positions for yourself: the propblem
            generation method is responsible to call you with all the possible
            values of `l`. You may however compare l and k to test if you have
            reached the horizon of your bounded problem generation.

        :param enc: the encoding used to store and organize the variables (used
            ie to shift vars)
        :param i: the time at which the generated expression will be considered
        :param k: the last time that exists in the universe of this expression
        :param l: l is the time position where the loop starts.
        :return: a boolean expression conform to the ltl bounded semantics of
            this node when there is no loop on the path from i to k
        """
        pass

    def nnf(self, negated):
        """
        All nodes of the AST must implement this function.
        Concretely, the role of this function is to return a version of `self`
        that is formatted according to the Negative Normal Form.

        :param negated: a flag indicating whether or not the subformula
            represented by self was negated in the parent formula
        :return: a nnf version of self
        """
        pass

class Atomic(Formula):
    """An abstract base class representing AST of an atomic proposition"""
    def __init__(self, x):
        self.id = x

    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.id)

class Unary(Formula):
    """An abstract base class representing AST of an unary proposition"""
    def __init__(self, prop):
        self.prop = prop

    def __repr__(self):
        return "({} {})".format(type(self).__name__, self.prop)

class Binary(Formula):
    """An abstract base class representing AST of a binary proposition"""
    def __init__(self, lhs=None, rhs=None):
        self.lhs = lhs
        self.rhs = rhs

    def __repr__(self):
        return "({} {} {})".format(self.lhs, type(self).__name__, self.rhs)

###############################################################################
# Propositional logic ast nodes
###############################################################################
class Constant(Atomic):
    def semantic_no_loop(self, enc, i, k):
        if self.id == "TRUE":
            return Be.true(enc.manager)
        else:
            return Be.false(enc.manager)

    def semantic_with_loop(self, enc, i, k, l):
        return self.semantic_no_loop(enc, i, k)

    def nnf(self, negated):
        if not negated:
            return self
        else:
            return Constant("FALSE") if self.id == "TRUE" else Constant("TRUE")

class Proposition(Atomic):
    def __init__(self, x):
        super().__init__(x)
        self._booleanized = None
        
    def _booleanize(self):
        """
        Returns a boolean expression (Be) corresponding to this simple
        boolean expression (text).

        .. note::
            Albeit feasible, working directly with variables as offered by the
            encoding is a little bit limiting as it de facto rejects any symbol
            which is not a variable. As a consequence, the DEFINES, or
            arithmetic expressions are not usable.

            The use of this function palliates that limitation and makes the
            use of any simple boolean expression possible.

        :return: a be expression (Be) corresponding to `self`
        """
        if not self._booleanized:
            befsm = master_be_fsm()
            node  = parse_simple_expression(self.id)
            self._booleanized = Wff(node).to_boolean_wff().to_be(befsm.encoding)
        return self._booleanized

    def _at_time(self, time):
        """
        Returns a boolean expression (Be) corresponding to this simple
        boolean expression (text) at the time step `time`.

        .. note::
            Albeit feasible, working directly with variables as offered by the
            encoding is a little bit limiting as it de facto rejects any symbol
            which is not a variable. As a consequence, the DEFINES, or
            arithmetic expressions are not usable.

            The use of this function palliates that limitation and makes the
            use of any simple boolean expression possible.

        :param time: the time at which the symbol should be shifted.
        :return: a be expression (Be) corresponding to `self` at `time`
        """
        booleanized = self._booleanize()
        return master_be_fsm().encoding.shift_to_time(booleanized, time)

    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        return self._at_time(i)

    def semantic_with_loop(self, enc, i, k, l):
        return self.semantic_no_loop(enc, i, k)

    def nnf(self, negated):
        return self if not negated else Not(self)

class Not(Unary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        return - self.prop.semantic_no_loop(enc, i, k)

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        return -self.prop.semantic_with_loop(enc, i, k, l)

    def nnf(self, negated):
        # double negation removes itself altogether
        return self.prop.nnf(True) if not negated else self.prop.nnf(False)

class And(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        lhs = self.lhs.semantic_no_loop(enc, i, k)
        rhs = self.rhs.semantic_no_loop(enc, i, k)
        return lhs & rhs

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        lhs = self.lhs.semantic_with_loop(enc, i, k, l)
        rhs = self.rhs.semantic_with_loop(enc, i, k, l)
        return lhs & rhs

    def nnf(self, negated):
        if not negated:
            return And(self.lhs.nnf(False), self.rhs.nnf(False))
        else:
            return Or(self.lhs.nnf(True), self.rhs.nnf(True))


class Or(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        lhs = self.lhs.semantic_no_loop(enc, i, k)
        rhs = self.rhs.semantic_no_loop(enc, i, k)
        return lhs | rhs

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        lhs = self.lhs.semantic_with_loop(enc, i, k, l)
        rhs = self.rhs.semantic_with_loop(enc, i, k, l)
        return lhs | rhs

    def nnf(self, negated):
        if not negated:
            return Or(self.lhs.nnf(False), self.rhs.nnf(False))
        else:
            return And(self.lhs.nnf(True), self.rhs.nnf(True))

class Xor(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        lhs = self.lhs.semantic_no_loop(enc, i, k)
        rhs = self.rhs.semantic_no_loop(enc, i, k)
        return lhs ^ rhs

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        lhs = self.lhs.semantic_with_loop(enc, i, k, l)
        rhs = self.rhs.semantic_with_loop(enc, i, k, l)
        return lhs ^ rhs

    def nnf(self, negated):
        if not negated:
            return Xor(self.lhs.nnf(False), self.rhs.nnf(False))
        else:
            # rewrite using : p ^ q <=> (p | q) & !(p & q)
            rewrite = And(Or(self.lhs, self.rhs), Not(And(self.lhs, self.rhs)))
            return rewrite.nnf(negated)

class Imply(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        lhs = self.lhs.semantic_no_loop(enc, i, k)
        rhs = self.rhs.semantic_no_loop(enc, i, k)
        return lhs.imply(rhs)

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        lhs = self.lhs.semantic_with_loop(enc, i, k, l)
        rhs = self.rhs.semantic_with_loop(enc, i, k, l)
        return lhs.imply(rhs)

    def nnf(self, negated):
        return Or(Not(self.lhs), self.rhs).nnf(negated)

class Equiv(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        lhs = self.lhs.semantic_no_loop(enc, i, k)
        rhs = self.rhs.semantic_no_loop(enc, i, k)
        return lhs.iff(rhs)

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        lhs = self.lhs.semantic_with_loop(enc, i, k, l)
        rhs = self.rhs.semantic_with_loop(enc, i, k, l)
        return lhs.iff(rhs)

    def nnf(self, negated):
        return And(Imply(self.lhs, self.rhs), Imply(self.rhs, self.lhs)).nnf(negated)

###############################################################################
# LTL specific ast nodes
###############################################################################

class Until(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        """The semantics when there is no loop:: [[lhs U rhs]]_{bound}^{time}"""
        # k is not infinity when there is no loop
        if i > k:
            return Be.false(enc.manager)

        psi = self.rhs.semantic_no_loop(enc, i, k)
        phi = self.lhs.semantic_no_loop(enc, i, k)
        return psi | (phi & self.semantic_no_loop(enc, i+1, k))

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        """The semantics when there is a loop:: _{l}[[lhs U rhs]]_{bound}^{time}"""
        # without moving at least one step, it is impossible to go through a loop
        if k == 0:
            return Be.false(enc.manager)

        def _semantic(time, cnt):
            """auxiliary function to stop recursing after k steps"""
            # at infinity, it is false: psi MUST happen at some time
            if cnt == k:
                return Be.false(enc.manager)
            psi = self.rhs.semantic_with_loop(enc, time, k, l)
            phi = self.lhs.semantic_with_loop(enc, time, k, l)
            return psi | (phi & _semantic(successor(time, k, l), cnt+1))

        return _semantic(i, 0)

    def nnf(self, negated):
        if not negated:
            return Until(self.lhs.nnf(False), self.rhs.nnf(False))
        else:
            # pseudo duality rule
            phi = self.lhs.nnf(True)
            psi = self.rhs.nnf(True)
            return WeakUntil(psi, And(phi, psi))

class WeakUntil(Binary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        """The semantics when there is no loop:: [[lhs W rhs]]_{bound}^{time}"""
        # k is not infinity when there is no loop
        if i > k:
            return Be.false(enc.manager)

        psi = self.rhs.semantic_no_loop(enc, i, k)
        phi = self.lhs.semantic_no_loop(enc, i, k)
        return psi | (phi & self.semantic_no_loop(enc, i+1, k))

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        """The semantics when there is a loop:: _{l}[[lhs W rhs]]_{bound}^{time}"""
        # without moving at least one step, it is impossible to go through a loop
        if k == 0:
            return Be.false(enc.manager)

        def _semantic(time, cnt):
            """auxiliary function to stop recursing after k steps"""
            # at infinity, it is true: psi is not forced if []phi
            if cnt == k:
                return Be.true(enc.manager)
            psi = self.rhs.semantic_with_loop(enc, time, k, l)
            phi = self.lhs.semantic_with_loop(enc, time, k, l)
            return psi | (phi & _semantic(successor(time, k, l), cnt+1))

        return _semantic(i, 0)

    def nnf(self, negated):
        if not negated:
            return WeakUntil(self.lhs.nnf(False), self.rhs.nnf(False))
        else:
            # pseudo duality rule
            phi = self.lhs.nnf(True)
            psi = self.rhs.nnf(True)
            return Until(psi, And(phi, psi))

class Globally(Unary):
    def semantic_no_loop(self, enc, i, k):
        return Be.false(enc.manager)
    
    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        # without moving at least one step, it is impossible to go through a loop
        if k == 0:
            return Be.false(enc.manager)

        def _semantic(time, cnt):
            if cnt == k:
                return Be.true(enc.manager)
            now = self.prop.semantic_with_loop(enc, time, k, l)
            return now & _semantic(successor(time, k, l), cnt+1)

        return _semantic(i, 0)

    def nnf(self, negated):
        if not negated:
            return Globally(self.prop.nnf(False))
        else:
            return Eventually(self.prop.nnf(True))

class Eventually(Unary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        if i > k:
            return Be.false(enc.manager)
        
        now = self.prop.semantic_no_loop(enc, i, k)
        return now | self.semantic_no_loop(enc, i+1, k)

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        # without moving at least one step, it is impossible to go through a loop
        if k == 0:
            return Be.false(enc.manager)
        
        def _semantic(time, cnt):
            if cnt == k:
                return Be.false(enc.manager)
            now = self.prop.semantic_with_loop(enc, time, k, l)
            return now | _semantic(successor(time, k, l), cnt+1)
        
        return _semantic(i, 0)

    def nnf(self, negated):
        if not negated:
            return Eventually(self.prop.nnf(False))
        else:
            return Globally(self.prop.nnf(True))

class Next(Unary):
    @memoize_no_loop
    def semantic_no_loop(self, enc, i, k):
        if i >= k:
            return Be.false(enc.manager)
        else:
            return self.prop.semantic_no_loop(enc, i+1, k)

    @memoize_with_loop
    def semantic_with_loop(self, enc, i, k, l):
        # without moving at least one step, it is impossible to go through a loop
        if k == 0:
            return Be.false(enc.manager)

        return self.prop.semantic_with_loop(enc, successor(i, k, l), k, l)

    def nnf(self, negated):
        return Next(self.prop.nnf(negated))
        



'''
This module validates the behavior of the functions composing the tool verifying
the diagnosability of some property using bounded model checking.
'''

from unittest                  import TestCase
from tests                     import utils as tests
                               
from pynusmv.init              import init_nusmv, deinit_nusmv
from pynusmv.glob              import load 
from pynusmv.bmc.glob          import go_bmc, bmc_exit, master_be_fsm
                               
from pynusmv.be.expression     import Be
from pynusmv.bmc               import ltlspec, utils as bmcutils

from pynusmv.node              import Node 
from pynusmv.parser            import parse_simple_expression, parse_ltl_spec 
from pynusmv.sat               import SatSolverFactory, SatSolverResult, Polarity
from pynusmv.wff               import Wff
from pynusmv_tools             import diagnosability

class TestDiagnosability(TestCase):
    
    def model(self):
        return tests.current_directory(__file__)+"/input.smv"
    
    def setUp(self):
        init_nusmv()
        load(self.model())
        go_bmc()
        
    def tearDown(self):
        bmc_exit()
        deinit_nusmv()
    
    def test_generate_path_only_valid_offset_params(self):
        # must combine an initial state w/ 'length' unrolling
        with self.assertRaises(ValueError):
            diagnosability.generate_path(-1, 5)
        
        with self.assertRaises(ValueError):
            diagnosability.generate_path(0, -5)
            
    def test_generate_path_offset_zero(self):
        # when offset is 0 path is just your regular path
        
        # - test 1 - just the initial state
        regular = bmcutils.BmcModel().path(0)
        tool    = diagnosability.generate_path(0, 0)
        self.assertEqual(regular, tool)
        
        # - test 2 - more steps unrolled
        regular = bmcutils.BmcModel().path(3)
        tool    = diagnosability.generate_path(0, 3)
        self.assertEqual(regular, tool)
        
    def test_generate_path_offset_x(self):
        # must combine an initial state w/ 'length' unrolling
        model   = bmcutils.BmcModel()
        manual  = model.init[5] & model.unrolling(5, 10)
        tool    = diagnosability.generate_path(5, 5)
        self.assertEqual(manual, tool)
    
    def test_constraint_same_observations(self):
        observable = diagnosability.mk_observable_vars(["mouse"])
        constraint = diagnosability.constraint_same_observations(observable, 0, 5, 5)
        model      = bmcutils.BmcModel()
        
        manual     = Be.true(model._fsm.encoding.manager)
        for i in range(6): # because we want to go from 0 through 5
            for v in model._fsm.encoding.input_variables:
                v_1 = v.at_time[i].boolean_expression
                v_2 = v.at_time[5+i].boolean_expression
                manual &= v_1.iff(v_2)
        
        self.assertEqual(manual, constraint)
        
    def test_eventually_critical_pair(self):
        enc= master_be_fsm().encoding
        f1 = Node.from_ptr(parse_simple_expression("status = active"))
        f2 = Node.from_ptr(parse_simple_expression("status = highlight"))
        
        constraint = diagnosability.constraint_eventually_critical_pair((f1, f2), 0, 5, 5)
        
        nnf1   = bmcutils.make_nnf_boolean_wff(f1).to_be(enc)
        nnf2   = bmcutils.make_nnf_boolean_wff(f2).to_be(enc)
        manual = Be.false(enc.manager)
        
        for i in range(6): # again, from 0 to 5
            manual |= ( enc.shift_to_time(nnf1 , i)
                      & enc.shift_to_time(nnf2 , 5+i))
        
        # observing the clauses generated in both cases, one observes that 
        # the generated clauses are the same except that the number of the cnf
        # literals do not match, example:
        #                        [-59, -24, 58]
        #                        [-65, -24, 64]
        # This is due to the fact that some 'fresh' cnf literals are used in the
        # generation of the epxression. Therefore, a comparison (even on the 
        # canonical form of the CNF) is not feasible.
        #
        # Satisfiability is just an indication but at least that is .. something
        solver_c = SatSolverFactory.create()
        cnf      = constraint.to_cnf()
        solver_c+= cnf
        solver_c.polarity(cnf, Polarity.POSITIVE)
        result_c = solver_c.solve()
        
        solver_m = SatSolverFactory.create()
        cnf      = manual.to_cnf()
        solver_m+= cnf
        solver_m.polarity(cnf, Polarity.POSITIVE)
        result_m = solver_m.solve()
        
        self.assertEqual(result_c, result_m)
        
    def test_generate_sat_problem(self):
        theta = Node.from_ptr(parse_simple_expression("TRUE"))
        theta = bmcutils.make_nnf_boolean_wff(theta)
        
        sigma_12= Node.from_ptr(parse_ltl_spec("TRUE"))
        sigma_12= bmcutils.make_nnf_boolean_wff(sigma_12).to_node()
        
        observable = diagnosability.mk_observable_vars(["mouse"])
        f1 = Node.from_ptr(parse_simple_expression("status = active"))
        f2 = Node.from_ptr(parse_simple_expression("status = inactive"))
         
        for i in range(5):
            problem = diagnosability.generate_sat_problem(observable, (f1, f2), i, theta, sigma_12, sigma_12)
            solver  = SatSolverFactory.create()
            cnf     = problem.to_cnf()
            solver += cnf
            solver.polarity(cnf, Polarity.POSITIVE)
            self.assertEqual(SatSolverResult.UNSATISFIABLE, solver.solve())
             
        f1 = Node.from_ptr(parse_simple_expression("status = active"))
        f2 = Node.from_ptr(parse_simple_expression("status = highlight"))
         
        for i in range(1, 4): 
            # length zero has no input => only an initial state and the 
            # diagnosability condition is not checked
            problem = diagnosability.generate_sat_problem(observable, (f1, f2), i, theta, sigma_12, sigma_12)
            solver  = SatSolverFactory.create()
            cnf     = problem.to_cnf()
            solver += cnf
            solver.polarity(cnf, Polarity.POSITIVE)
            self.assertEqual(SatSolverResult.SATISFIABLE, solver.solve())

    def test_verify_exactly(self):
        theta = Node.from_ptr(parse_simple_expression("TRUE"))
        theta = bmcutils.make_nnf_boolean_wff(theta)
        
        sigma_12= Node.from_ptr(parse_ltl_spec("TRUE"))
        sigma_12= bmcutils.make_nnf_boolean_wff(sigma_12).to_node()
        
        obs_names = ["mouse"]
        obs_vars  = diagnosability.mk_observable_vars(obs_names)
        f1 = Node.from_ptr(parse_simple_expression("status = active"))
        f2 = Node.from_ptr(parse_simple_expression("status = inactive"))
        
        for i in range(5):
            res = diagnosability.verify_for_size_exactly_k(obs_names, obs_vars, (f1, f2), i, theta, sigma_12, sigma_12)
            self.assertEqual("No Violation", res)
        
        f1 = Node.from_ptr(parse_simple_expression("status = active"))
        f2 = Node.from_ptr(parse_simple_expression("status = highlight"))
        
        res = diagnosability.verify_for_size_exactly_k(obs_names, obs_vars, (f1, f2), 0, theta, sigma_12, sigma_12)
        self.assertEqual("No Violation", res)
        
        res = diagnosability.verify_for_size_exactly_k(obs_names, obs_vars, (f1, f2), 1, theta, sigma_12, sigma_12)
        self.assertTrue(res.startswith("############### DIAGNOSABILITY VIOLATION"))
        
        res = diagnosability.verify_for_size_exactly_k(obs_names, obs_vars, (f1, f2), 2, theta, sigma_12, sigma_12)
        self.assertTrue(res.startswith("############### DIAGNOSABILITY VIOLATION"))
        
        res = diagnosability.verify_for_size_exactly_k(obs_names, obs_vars, (f1, f2), 3, theta, sigma_12, sigma_12)
        self.assertTrue(res.startswith("############### DIAGNOSABILITY VIOLATION"))
        
    def test_mk_observable_vars(self):
        enc = master_be_fsm().encoding
        
        # if a non existing var name is passed, an exception is thrown
        with self.assertRaises(ValueError):
                diagnosability.mk_observable_vars(["a"])
                
        observable = diagnosability.mk_observable_vars(["status"])
        self.assertEqual(observable, [enc.by_name["status.1"], enc.by_name["status.0"]])
        
    def test_constraint_context_theta(self):
        enc   = master_be_fsm().encoding
        cond  = Wff(parse_simple_expression("mouse = down")).to_boolean_wff()
        
        theta = diagnosability.constraint_context_theta_initial(cond, 0, 1)
        manual= enc.shift_to_time(cond.to_be(enc), 0) \
              & enc.shift_to_time(cond.to_be(enc), 1) \
              
        self.assertEqual(theta, manual)
        
    def test_constraint_context_sigma(self):
        fsm   = master_be_fsm()
        
        _true = Node.from_ptr(parse_ltl_spec("TRUE"))
        _true = bmcutils.make_nnf_boolean_wff(_true)
        _truen= _true.to_node()
        
        cond  = Wff(parse_ltl_spec("G !(mouse = hover)"))\
                    .to_boolean_wff()\
                    .to_negation_normal_form()
        off_1 = 0
        off_2 = 2
        length= 1
         
        # sigma1
        problem = diagnosability.generate_sat_problem([], (_truen, _truen), length, _true, cond.to_node(), _truen)
        tm_cond = ltlspec.bounded_semantics_at_offset(fsm, cond.to_node(), length, off_1)
        
        canonical_p = tests.canonical_cnf(problem)
        canonical_f = tests.canonical_cnf(tm_cond)
        
        self.assertTrue(all(clause in canonical_p for clause in canonical_f))
        
        # sigma2
        problem = diagnosability.generate_sat_problem([], (_truen, _truen), length, _true, _truen, cond.to_node())
        tm_cond = ltlspec.bounded_semantics_at_offset(fsm, cond.to_node(), length, off_2)
        
        canonical_p = tests.canonical_cnf(problem)
        canonical_f = tests.canonical_cnf(tm_cond)
        
        self.assertTrue(all(clause in canonical_p for clause in canonical_f))
        


import sys
import argparse

from pynusmv.prop import PropDb
from pynusmv.prop import propTypes
from pynusmv.init import init_nusmv, deinit_nusmv
from pynusmv import glob
from pynusmv.fsm import BddFsm

from pynusmv_tools.tlace.check import check as check_ctl_spec
from pynusmv_tools.tlace.xml import xml_representation
    
    
def check_and_explain(allargs):
    """
    Check specs on the given NuSMV model and compute TLACEs when needed.
    
    Build the model from a given file, check every CTL spec in it
    and compute and store TLACEs when needed.
    
    allargs -- a sys.args-like arguments list, without script name.
    """
    
    # Parse arguments
    parser = argparse.ArgumentParser(description='CTL model checker '
                                                 'with TLACE generation.')
    # Populate arguments: for now, only the model
    parser.add_argument('model', help='the NuSMV model with specifications')
    args = parser.parse_args(allargs)
    
    # Initialize the model
    fsm = BddFsm.from_filename(args.model)
    propDb = glob.prop_database()
    
    # Check all CTL properties
    for prop in propDb:
        #  Check type
        if prop.type == propTypes['CTL']:
            spec = prop.exprcore
    
            (satisfied, cntex) = check_ctl_spec(fsm, spec)
            # Print the result and the TLACE if any
            print('Specification',str(spec), 'is', str(satisfied),
                  file=sys.stderr)
        
            if not satisfied:
                print(xml_representation(fsm, cntex, spec))
            
            print()

def main():
    '''TLACE entry point'''
    # Initialize/Quit NuSMV
    with init_nusmv():
        check_and_explain(sys.argv[1:])

if __name__ == "__main__":
    main()



class Tlacebranch:
    """
    A Tlacebranch is a TLACE branch.
    It contains an existential formula and a TLACE path,
    represented by a tuple of a list of TLACE nodes
    and a possibly None looping node.
    """
    
    def __init__(self, formula, path, loop=None):
        """
        Create a new TLACE branch.
        
        formula -- an existential temporal formula;
        path -- a list of TLACE nodes;
        loop -- None or a (inputs, node) pair where node is a node of path
                indicating the start of a loop.
        """
        self.__formula = formula
        self.__path = (path, loop)
        
        
    def __str__(self):
        pass # TODO
        
    
    @property
    def specification(self):
        """The spec of this branch."""
        return self.__formula
        
    @property
    def path(self):
        """The path of this branch. A (path, (inputs, loop)) tuple, or None."""
        return self.__path